@@ -1,11 +1,11 @@
-# -*- coding: utf-8 -*-
 # -*- mode: python -*-
 """spike detection and extraction
 
 Copyright (C) 2013 Dan Meliza <dmeliza@gmail.com>
 Created Fri Jul 12 14:05:16 2013
 """
-from typing import Iterable, Union, Tuple, Optional, Iterator, Mapping
+from typing import Callable, Iterable, Iterator, Optional, Tuple, Union
+
 import numpy as np
 
 numeric = Union[int, float, np.number]
@@ -17,62 +17,77 @@ def filter_times(
     return tuple(t for t in times if (t > min) and (t < max))
 
 
+def peak_idx(spikes: np.ndarray):
+    """Return index of maximum negative value"""
+    return spikes.argmax(-1)
+
+
+def trough_idx(spikes: np.ndarray):
+    """Return index of maximum positive value"""
+    return spikes.argmin(-1)
+
+
+def upswing_idx(spikes: np.ndarray):
+    """Return index of the point midway between the trough and the following peak"""
+    _trough = trough_idx(spikes)
+    # need to find the local maximum after the trough, not the global max
+    raise NotImplementedError
+
+
 def realign_spikes(
     times: Iterable[numeric],
     spikes: np.ndarray,
-    upsample: int,
+    *,
+    upsample: int = 1,
+    align_by: Callable[[np.ndarray], Union[np.ndarray, int]] = peak_idx,
     jitter: int = 3,
-    reflect_fft: bool = False,
     expected_peak: Optional[int] = None,
 ) -> Tuple[np.ndarray, np.ndarray]:
-    """Realign spikes to their peaks using bandwidth-limited resampling
+    """Realign spikes to their peaks (or some other feature) using bandwidth-limited resampling
 
     times    : one-dimensional array of spike times, in units of samples
     spikes   : array of spike waveforms, with dimensions (nspikes, npoints)
     upsample : integer, the factor by which to upsample the data for realignment (must be greater than 1)
-    jitter   : samples (in original data) to search for true peak
-    expected_peak : if supplied, searches around this index for the peak. If not supplied,
-                    uses the argmax of the mean spike
+    align_by : function that returns the alignment point in each spike (see below)
+    jitter   : samples (in original data) to search for the alignment point
+    expected_peak : if supplied, searches around this index (in original data) for the peak. If not supplied,
+                    uses align_by on the mean spike
 
-    Returns (times, spikes), with the sampling rate increased by a factor of upsample
+    The align_by function takes a 1D or 2D array and returns the index
+    corresponding to the alignment point in the spike. If the array is 1D, the
+    function should return a single number. If the array is 2D, the function
+    must return an array. For example, peak_idx is lambda x: x.argmax(-1)
+
+    Returns (times, spikes), with the sampling rate increased by a factor of
+    upsample.
 
     """
-    assert (
-        isinstance(upsample, int) and upsample > 1
-    ), "Upsampling factor must be an integer greater than 1"
+    if not (isinstance(upsample, int) and upsample > 0):
+        raise ValueError("Upsampling factor must be a positive integer")
     nevents, nsamples = spikes.shape
 
-    # first infer the expected peak time
     if expected_peak is None:
-        expected_peak = np.mean(spikes, 0).argmax() * upsample
-    else:
-        expected_peak *= upsample
-    spikes = fftresample(spikes, int(nsamples * upsample), reflect=reflect_fft)
+        # infer expected peak from the mean spike
+        expected_peak = align_by(np.mean(spikes, 0))
+    expected_peak *= upsample
+
+    if upsample > 1:
+        spikes = fftresample(spikes, int(nsamples * upsample), reflect=True)
     # find peaks within upsample samples of mean peak
-    shift = find_peaks(spikes, expected_peak, upsample * jitter)
-    start = shift + upsample * jitter
+    window = upsample * jitter
+    # output array is clipped by jitter on both ends and then upsampled
     nshifted = (nsamples - 2 * jitter) * upsample
     shifted = np.zeros((nevents, nshifted))
+    # offset is the position of the peak (or trough) within the search window
+    search = slice(expected_peak - window, expected_peak + window + 1)
+    start = align_by(spikes[:, search])
     for i, spike in enumerate(spikes):
         shifted[i, :] = spike[start[i] : start[i] + nshifted]
-    return (np.asarray(times) * upsample + shift, shifted)
+    # spike times are shifted by offset - search window
+    return (np.asarray(times) * upsample + start - window, shifted)
 
 
-def find_peaks(spikes: np.ndarray, peak: int, window: int) -> np.ndarray:
-    """Locate the peaks in an array of spikes.
-
-    spikes: resampled spike waveforms, dimensions (nspikes, nsamples)
-    peak:   the expected peak location
-    window: the number of samples to either side of the peak to look for the peak
-
-    Returns array of shift values relative to peak
-
-    """
-    r = slice(peak - window, peak + window + 1)
-    return spikes[:, r].argmax(1) - window
-
-
-def fftresample(S: np.ndarray, npoints: int, reflect: bool = False) -> np.ndarray:
+def fftresample(S: np.ndarray, npoints: int, *, reflect: bool = False) -> np.ndarray:
     """Resample an array of waveforms using discrete fourier transform. The
     signal is transformed in the fourier domain and then padded or truncated to
     the desired size. This should be equivalent to a sinc resampling.
@@ -82,7 +97,7 @@ def fftresample(S: np.ndarray, npoints: int, reflect: bool = False) -> np.ndarra
     reflect: if True, pad the sample with reflected copies on either end to suppress edge effects
 
     """
-    from numpy.fft import rfft, irfft
+    from numpy.fft import irfft, rfft
 
     if reflect:
         Srev = S[:, ::-1]
@@ -98,7 +113,7 @@ def fftresample(S: np.ndarray, npoints: int, reflect: bool = False) -> np.ndarra
 
 
 def trim_waveforms(
-    spikes: Iterable[np.ndarray], times: Iterable[int], peak_t: int, n_rise: int
+    spikes: Iterable[np.ndarray], times: Iterable[int], *, peak_t: int, n_rise: int
 ) -> Iterator[Tuple[int, np.ndarray]]:
     """Trims spike waveforms to remove overlapping peaks.
 
